package org.apache.bcel.classfile;

import java.util.Objects;
import java.util.List;
import java.util.ArrayList;
import org.apache.bcel.generic.Type;
import java.io.ByteArrayOutputStream;
import java.util.Set;
import java.util.TreeSet;
import org.apache.bcel.util.ClassQueue;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.DataOutputStream;
import java.util.Arrays;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.bcel.util.SyntheticRepository;
import java.util.StringTokenizer;
import org.apache.bcel.util.Repository;
import org.apache.bcel.util.BCELComparator;

public class JavaClass extends AccessFlags implements Cloneable, Node, Comparable<JavaClass>
{
    public static final JavaClass[] EMPTY_ARRAY;
    public static final byte HEAP = 1;
    public static final byte FILE = 2;
    public static final byte ZIP = 3;
    private static final boolean debug;
    private static BCELComparator bcelComparator;
    private String fileName;
    private final String packageName;
    private String sourceFileName;
    private int classNameIndex;
    private int superclassNameIndex;
    private String className;
    private String superclassName;
    private int major;
    private int minor;
    private ConstantPool constantPool;
    private int[] interfaces;
    private String[] interfaceNames;
    private Field[] fields;
    private Method[] methods;
    private Attribute[] attributes;
    private AnnotationEntry[] annotations;
    private byte source;
    private boolean isAnonymous;
    private boolean isNested;
    private boolean computedNestedTypeStatus;
    private transient Repository repository;
    
    static void Debug(final String str) {
        if (JavaClass.debug) {
            System.out.println(str);
        }
    }
    
    public static BCELComparator getComparator() {
        return JavaClass.bcelComparator;
    }
    
    private static String indent(final Object obj) {
        final StringTokenizer tokenizer = new StringTokenizer(obj.toString(), "\n");
        final StringBuilder buf = new StringBuilder();
        while (tokenizer.hasMoreTokens()) {
            buf.append("\t").append(tokenizer.nextToken()).append("\n");
        }
        return buf.toString();
    }
    
    public static void setComparator(final BCELComparator comparator) {
        JavaClass.bcelComparator = comparator;
    }
    
    public JavaClass(final int classNameIndex, final int superclassNameIndex, final String fileName, final int major, final int minor, final int accessFlags, final ConstantPool constantPool, final int[] interfaces, final Field[] fields, final Method[] methods, final Attribute[] attributes) {
        this(classNameIndex, superclassNameIndex, fileName, major, minor, accessFlags, constantPool, interfaces, fields, methods, attributes, (byte)1);
    }
    
    public JavaClass(final int classNameIndex, final int superclassNameIndex, final String fileName, final int major, final int minor, final int accessFlags, final ConstantPool constantPool, int[] interfaces, Field[] fields, Method[] methods, Attribute[] attributes, final byte source) {
        super(accessFlags);
        this.sourceFileName = "<Unknown>";
        this.source = 1;
        this.repository = SyntheticRepository.getInstance();
        if (interfaces == null) {
            interfaces = ArrayUtils.EMPTY_INT_ARRAY;
        }
        if (attributes == null) {
            attributes = Attribute.EMPTY_ARRAY;
        }
        if (fields == null) {
            fields = Field.EMPTY_FIELD_ARRAY;
        }
        if (methods == null) {
            methods = Method.EMPTY_METHOD_ARRAY;
        }
        this.classNameIndex = classNameIndex;
        this.superclassNameIndex = superclassNameIndex;
        this.fileName = fileName;
        this.major = major;
        this.minor = minor;
        this.constantPool = constantPool;
        this.interfaces = interfaces;
        this.fields = fields;
        this.methods = methods;
        this.attributes = attributes;
        this.source = source;
        for (final Attribute attribute : attributes) {
            if (attribute instanceof SourceFile) {
                this.sourceFileName = ((SourceFile)attribute).getSourceFileName();
                break;
            }
        }
        this.className = constantPool.getConstantString(classNameIndex, (byte)7);
        this.className = Utility.compactClassName(this.className, false);
        final int index = this.className.lastIndexOf(46);
        if (index < 0) {
            this.packageName = "";
        }
        else {
            this.packageName = this.className.substring(0, index);
        }
        if (superclassNameIndex > 0) {
            this.superclassName = constantPool.getConstantString(superclassNameIndex, (byte)7);
            this.superclassName = Utility.compactClassName(this.superclassName, false);
        }
        else {
            this.superclassName = "java.lang.Object";
        }
        this.interfaceNames = new String[interfaces.length];
        for (int i = 0; i < interfaces.length; ++i) {
            final String str = constantPool.getConstantString(interfaces[i], (byte)7);
            this.interfaceNames[i] = Utility.compactClassName(str, false);
        }
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitJavaClass(this);
    }
    
    @Override
    public int compareTo(final JavaClass obj) {
        return this.getClassName().compareTo(obj.getClassName());
    }
    
    private void computeNestedTypeStatus() {
        if (this.computedNestedTypeStatus) {
            return;
        }
        for (final Attribute attribute : this.attributes) {
            if (attribute instanceof InnerClasses) {
                ((InnerClasses)attribute).forEach(innerClass -> {
                    boolean innerClassAttributeRefersToMe = false;
                    final String innerClassName = this.constantPool.getConstantString(innerClass.getInnerClassIndex(), (byte)7);
                    final String innerClassName2 = Utility.compactClassName(innerClassName, false);
                    if (innerClassName2.equals(this.getClassName())) {
                        innerClassAttributeRefersToMe = true;
                    }
                    if (innerClassAttributeRefersToMe) {
                        this.isNested = true;
                        if (innerClass.getInnerNameIndex() == 0) {
                            this.isAnonymous = true;
                        }
                    }
                    return;
                });
            }
        }
        this.computedNestedTypeStatus = true;
    }
    
    public JavaClass copy() {
        try {
            final JavaClass c = (JavaClass)this.clone();
            c.constantPool = this.constantPool.copy();
            c.interfaces = this.interfaces.clone();
            c.interfaceNames = this.interfaceNames.clone();
            Arrays.setAll(c.fields = new Field[this.fields.length], i -> this.fields[i].copy(c.constantPool));
            Arrays.setAll(c.methods = new Method[this.methods.length], i -> this.methods[i].copy(c.constantPool));
            Arrays.setAll(c.attributes = new Attribute[this.attributes.length], i -> this.attributes[i].copy(c.constantPool));
            return c;
        }
        catch (final CloneNotSupportedException e) {
            return null;
        }
    }
    
    public void dump(final DataOutputStream file) throws IOException {
        file.writeInt(-889275714);
        file.writeShort(this.minor);
        file.writeShort(this.major);
        this.constantPool.dump(file);
        file.writeShort(super.getAccessFlags());
        file.writeShort(this.classNameIndex);
        file.writeShort(this.superclassNameIndex);
        file.writeShort(this.interfaces.length);
        for (final int interface1 : this.interfaces) {
            file.writeShort(interface1);
        }
        file.writeShort(this.fields.length);
        for (final Field field : this.fields) {
            field.dump(file);
        }
        file.writeShort(this.methods.length);
        for (final Method method : this.methods) {
            method.dump(file);
        }
        if (this.attributes != null) {
            file.writeShort(this.attributes.length);
            for (final Attribute attribute : this.attributes) {
                attribute.dump(file);
            }
        }
        else {
            file.writeShort(0);
        }
        file.flush();
    }
    
    public void dump(final File file) throws IOException {
        final String parent = file.getParent();
        if (parent != null) {
            final File dir = new File(parent);
            if (!dir.mkdirs() && !dir.isDirectory()) {
                throw new IOException("Could not create the directory " + dir);
            }
        }
        try (DataOutputStream dos = new DataOutputStream(new FileOutputStream(file))) {
            this.dump(dos);
        }
    }
    
    public void dump(final OutputStream file) throws IOException {
        this.dump(new DataOutputStream(file));
    }
    
    public void dump(final String fileName) throws IOException {
        this.dump(new File(fileName));
    }
    
    @Override
    public boolean equals(final Object obj) {
        return JavaClass.bcelComparator.equals(this, obj);
    }
    
    public JavaClass[] getAllInterfaces() throws ClassNotFoundException {
        final ClassQueue queue = new ClassQueue();
        final Set<JavaClass> allInterfaces = new TreeSet<JavaClass>();
        queue.enqueue(this);
        while (!queue.empty()) {
            final JavaClass clazz = queue.dequeue();
            final JavaClass souper = clazz.getSuperClass();
            final JavaClass[] interfaces = clazz.getInterfaces();
            if (clazz.isInterface()) {
                allInterfaces.add(clazz);
            }
            else if (souper != null) {
                queue.enqueue(souper);
            }
            for (final JavaClass iface : interfaces) {
                queue.enqueue(iface);
            }
        }
        return allInterfaces.toArray(JavaClass.EMPTY_ARRAY);
    }
    
    public AnnotationEntry[] getAnnotationEntries() {
        if (this.annotations == null) {
            this.annotations = AnnotationEntry.createAnnotationEntries(this.getAttributes());
        }
        return this.annotations;
    }
    
    public Attribute[] getAttributes() {
        return this.attributes;
    }
    
    public byte[] getBytes() {
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try (DataOutputStream dos = new DataOutputStream(baos)) {
            this.dump(dos);
        }
        catch (final IOException e) {
            e.printStackTrace();
        }
        return baos.toByteArray();
    }
    
    public String getClassName() {
        return this.className;
    }
    
    public int getClassNameIndex() {
        return this.classNameIndex;
    }
    
    public ConstantPool getConstantPool() {
        return this.constantPool;
    }
    
    public Field[] getFields() {
        return this.fields;
    }
    
    public String getFileName() {
        return this.fileName;
    }
    
    public int[] getInterfaceIndices() {
        return this.interfaces;
    }
    
    public String[] getInterfaceNames() {
        return this.interfaceNames;
    }
    
    public JavaClass[] getInterfaces() throws ClassNotFoundException {
        final String[] interfaces = this.getInterfaceNames();
        final JavaClass[] classes = new JavaClass[interfaces.length];
        for (int i = 0; i < interfaces.length; ++i) {
            classes[i] = this.repository.loadClass(interfaces[i]);
        }
        return classes;
    }
    
    public int getMajor() {
        return this.major;
    }
    
    public Method getMethod(final java.lang.reflect.Method m) {
        for (final Method method : this.methods) {
            if (m.getName().equals(method.getName()) && m.getModifiers() == method.getModifiers() && Type.getSignature(m).equals(method.getSignature())) {
                return method;
            }
        }
        return null;
    }
    
    public Method[] getMethods() {
        return this.methods;
    }
    
    public int getMinor() {
        return this.minor;
    }
    
    public String getPackageName() {
        return this.packageName;
    }
    
    public Repository getRepository() {
        return this.repository;
    }
    
    public final byte getSource() {
        return this.source;
    }
    
    public String getSourceFileName() {
        return this.sourceFileName;
    }
    
    public JavaClass getSuperClass() throws ClassNotFoundException {
        if ("java.lang.Object".equals(this.getClassName())) {
            return null;
        }
        return this.repository.loadClass(this.getSuperclassName());
    }
    
    public JavaClass[] getSuperClasses() throws ClassNotFoundException {
        JavaClass clazz = this;
        final List<JavaClass> allSuperClasses = new ArrayList<JavaClass>();
        for (clazz = clazz.getSuperClass(); clazz != null; clazz = clazz.getSuperClass()) {
            allSuperClasses.add(clazz);
        }
        return allSuperClasses.toArray(JavaClass.EMPTY_ARRAY);
    }
    
    public String getSuperclassName() {
        return this.superclassName;
    }
    
    public int getSuperclassNameIndex() {
        return this.superclassNameIndex;
    }
    
    @Override
    public int hashCode() {
        return JavaClass.bcelComparator.hashCode(this);
    }
    
    public boolean implementationOf(final JavaClass inter) throws ClassNotFoundException {
        if (!inter.isInterface()) {
            throw new IllegalArgumentException(inter.getClassName() + " is no interface");
        }
        if (this.equals(inter)) {
            return true;
        }
        final JavaClass[] allInterfaces;
        final JavaClass[] superInterfaces = allInterfaces = this.getAllInterfaces();
        for (final JavaClass superInterface : allInterfaces) {
            if (superInterface.equals(inter)) {
                return true;
            }
        }
        return false;
    }
    
    public final boolean instanceOf(final JavaClass superclass) throws ClassNotFoundException {
        if (this.equals(superclass)) {
            return true;
        }
        for (final JavaClass clazz : this.getSuperClasses()) {
            if (clazz.equals(superclass)) {
                return true;
            }
        }
        return superclass.isInterface() && this.implementationOf(superclass);
    }
    
    public final boolean isAnonymous() {
        this.computeNestedTypeStatus();
        return this.isAnonymous;
    }
    
    public final boolean isClass() {
        return (super.getAccessFlags() & 0x200) == 0x0;
    }
    
    public final boolean isNested() {
        this.computeNestedTypeStatus();
        return this.isNested;
    }
    
    public final boolean isSuper() {
        return (super.getAccessFlags() & 0x20) != 0x0;
    }
    
    public void setAttributes(final Attribute[] attributes) {
        this.attributes = attributes;
    }
    
    public void setClassName(final String className) {
        this.className = className;
    }
    
    public void setClassNameIndex(final int classNameIndex) {
        this.classNameIndex = classNameIndex;
    }
    
    public void setConstantPool(final ConstantPool constantPool) {
        this.constantPool = constantPool;
    }
    
    public void setFields(final Field[] fields) {
        this.fields = fields;
    }
    
    public void setFileName(final String fileName) {
        this.fileName = fileName;
    }
    
    public void setInterfaceNames(final String[] interfaceNames) {
        this.interfaceNames = interfaceNames;
    }
    
    public void setInterfaces(final int[] interfaces) {
        this.interfaces = interfaces;
    }
    
    public void setMajor(final int major) {
        this.major = major;
    }
    
    public void setMethods(final Method[] methods) {
        this.methods = methods;
    }
    
    public void setMinor(final int minor) {
        this.minor = minor;
    }
    
    public void setRepository(final Repository repository) {
        this.repository = repository;
    }
    
    public void setSourceFileName(final String sourceFileName) {
        this.sourceFileName = sourceFileName;
    }
    
    public void setSuperclassName(final String superclassName) {
        this.superclassName = superclassName;
    }
    
    public void setSuperclassNameIndex(final int superclassNameIndex) {
        this.superclassNameIndex = superclassNameIndex;
    }
    
    @Override
    public String toString() {
        String access = Utility.accessToString(super.getAccessFlags(), true);
        access = (access.isEmpty() ? "" : (access + " "));
        final StringBuilder buf = new StringBuilder(128);
        buf.append(access).append(Utility.classOrInterface(super.getAccessFlags())).append(" ").append(this.className).append(" extends ").append(Utility.compactClassName(this.superclassName, false)).append('\n');
        final int size = this.interfaces.length;
        if (size > 0) {
            buf.append("implements\t\t");
            for (int i = 0; i < size; ++i) {
                buf.append(this.interfaceNames[i]);
                if (i < size - 1) {
                    buf.append(", ");
                }
            }
            buf.append('\n');
        }
        buf.append("file name\t\t").append(this.fileName).append('\n');
        buf.append("compiled from\t\t").append(this.sourceFileName).append('\n');
        buf.append("compiler version\t").append(this.major).append(".").append(this.minor).append('\n');
        buf.append("access flags\t\t").append(super.getAccessFlags()).append('\n');
        buf.append("constant pool\t\t").append(this.constantPool.getLength()).append(" entries\n");
        buf.append("ACC_SUPER flag\t\t").append(this.isSuper()).append("\n");
        if (this.attributes.length > 0) {
            buf.append("\nAttribute(s):\n");
            for (final Attribute attribute : this.attributes) {
                buf.append(indent(attribute));
            }
        }
        final AnnotationEntry[] annotations = this.getAnnotationEntries();
        if (annotations != null && annotations.length > 0) {
            buf.append("\nAnnotation(s):\n");
            for (final AnnotationEntry annotation : annotations) {
                buf.append(indent(annotation));
            }
        }
        if (this.fields.length > 0) {
            buf.append("\n").append(this.fields.length).append(" fields:\n");
            for (final Field field : this.fields) {
                buf.append("\t").append(field).append('\n');
            }
        }
        if (this.methods.length > 0) {
            buf.append("\n").append(this.methods.length).append(" methods:\n");
            for (final Method method : this.methods) {
                buf.append("\t").append(method).append('\n');
            }
        }
        return buf.toString();
    }
    
    static {
        EMPTY_ARRAY = new JavaClass[0];
        debug = Boolean.getBoolean("JavaClass.debug");
        JavaClass.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(final Object o1, final Object o2) {
                final JavaClass THIS = (JavaClass)o1;
                final JavaClass THAT = (JavaClass)o2;
                return Objects.equals(THIS.getClassName(), THAT.getClassName());
            }
            
            @Override
            public int hashCode(final Object o) {
                final JavaClass THIS = (JavaClass)o;
                return THIS.getClassName().hashCode();
            }
        };
    }
}
